# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio import add_container, add_dataset, attach, close, delete_metadata, detach, erase, get_metadata, list_content, list_content_history, list_dids, list_parent_dids, reopen, set_metadata, stat, touch
from rucio.cli.utils import Arguments


@click.group()
def did():
    """Manage Data Identifiers - the source data objects"""


@did.command("list")
@click.option("-r", "--recursive", default=False, is_flag=True, help="List data identifiers recursively.")
@click.option(
    "--filter", "filter_",
    help="""
    Single or logically combined filtering expression(s) either in the form <key><operator><value>
    or <value1><operator1><key><operator2><value2> (compound inequality).
    Keys are equivalent to columns in the DID table.
    Operators must belong to the set of (<=, >=, ==, !=, >, <). The following conventions for combining expressions
    are used: ";" represents the logical OR operator; "," represents the logical AND operator',
    """,
)  # TODO Shorten this help and make supplying this easier
@click.option("--short", is_flag=True, default=False, help="Dump the list of DIDs")
@click.argument("did-pattern", nargs=-1)
@click.option("--parent", default=False, is_flag=True, help="List the parents of the DID - must use a full DID scope and name")
# TODO Implement or remove option - view https://github.com/rucio/rucio/issues/7230
@click.option("--pfn", hidden=True)
@click.option("--guid", hidden=True)
@click.pass_context
def list_(ctx, did_pattern, recursive, filter_, short, parent, pfn, guid):
    """
    List the Data IDentifiers matching certain pattern.
    Only the collections (i.e. dataset or container) are returned by default.
    With the filter option, you can specify a list of metadata that the Data IDentifier should match
    """
    if parent:
        for did in did_pattern:
            list_parent_dids(Arguments({"did": did}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    else:
        args = Arguments({"did": did_pattern, "recursive": recursive, "filter": filter_, "short": short})
        list_dids(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@did.command("show")
@click.argument("dids", nargs=-1)
@click.pass_context
def show(ctx, dids):
    """List attributes, statuses, or parents for data identifiers"""
    stat(Arguments({"dids": dids}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@did.command("add")
@click.argument("did-name")
@click.option("--type", "dtype", type=click.Choice(["container", "dataset"]))
@click.option("--monotonic", is_flag=True, default=False, help="Monotonic status to True.")
@click.option("--lifetime", type=int, help="Lifetime in seconds.")
@click.pass_context
def add_(ctx, did_name, dtype, monotonic, lifetime):
    """Create a new collection-type DID"""
    args = Arguments({"did": did_name, "monotonic": monotonic, "lifetime": lifetime})
    if dtype == "container":
        add_container(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    else:
        add_dataset(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@did.command("update")
@click.argument("dids", nargs=-1)
@click.option("--rse", "--rse-name", help="The RSE of the DIDs")
@click.option("--touch", "operation", flag_value="touch", default=True, help="Touch one or more DIDs and set the last accessed date to the current date")
@click.option("--open", "operation", flag_value="open", help="Reopen a dataset or container (only for privileged users)")
@click.option("--close", "operation", flag_value="close", help="Close a dataset or container.")
@click.pass_context
def update(ctx, dids, rse, operation):
    """Touch one or more DIDs and set the last accessed date to the current date, or mark them as open or closed."""
    args = Arguments({"dids": dids, "rse": rse})
    if operation == "touch":
        touch(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    elif operation == "open":
        reopen(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    elif operation == "close":
        close(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)
    else:
        raise ValueError("No operation specified, please use `--help` to see possibilities")  # Should not be possible, but better safe than sorry


@did.command("remove")
@click.option("--undo", is_flag=True, default=False, help="Undo erase DIDs. Only works if has been less than 24 hours since erase operation.")
@click.argument("dids", nargs=-1)
@click.pass_context
def remove(ctx, dids, undo):
    """
    This command sets the lifetime of the DID in order to expire in the next 24 hours.
    Expired DIDs are force-deleted (and their replicas purged).
    The deletion is not reversible after 24 hours grace time period expired
    """
    erase(Arguments({"dids": dids, "undo": undo}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@did.group()
def content():
    """Manage contents of collection type DIDs"""


@content.command("history")
@click.argument("dids", nargs=-1)
@click.pass_context
def content_history(ctx, dids):
    """List the content history of a collection-type DID"""
    list_content_history(Arguments({"dids": dids}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@content.command("add")
@click.option("-to", "--to-did", required=True, help="Collection-type DID to which attach [DIDs]")
@click.option("-f", "--from-file", is_flag=True, default=False, help="[DIDs] is a file instead of a list of did names. The file should contain one did per line.")
@click.argument("dids", nargs=-1)
@click.pass_context
def content_add_(ctx, to_did, from_file, dids):
    """Attach a list [dids] of Data IDentifiers (file or collection-type) to an other Data IDentifier (collection-type)"""
    args = Arguments({"dids": dids, "todid": to_did, "fromfile": from_file})
    attach(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@content.command("remove")
@click.option("-f", "--from-did", help="Collection-type DID to remove DIDs from")
@click.argument("dids", nargs=-1)
@click.pass_context
def content_remove(ctx, dids, from_did):
    """Detach [dids], a list of DIDs (file or collection-type) from an other Data Identifier (collection type)"""
    args = Arguments({"dids": dids, "fromdid": from_did})
    detach(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@content.command("list")
@click.argument("dids", nargs=-1)
@click.option("--short", is_flag=True, default=False, help="Just dump the list of DIDs.")
@click.pass_context
def content_list_(ctx, dids, short):
    """List the content of a collection-type DID"""
    args = Arguments({"dids": dids, "short": short})
    list_content(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@did.group()
def metadata():
    """Manage metadata for DIDs"""


@metadata.command("add")
@click.argument("did")
@click.option('--key', help='Attribute key', required=True)
@click.option('--value', help='Attribute value', required=True)
@click.pass_context
def metadata_add_(ctx, did, key, value):
    """Add metadata to a DID"""
    args = Arguments({"did": did, "key": key, "value": value})
    set_metadata(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@metadata.command("remove")
@click.argument("did")
@click.option("--key", help="Key to remove from a DID's metadata.")
@click.pass_context
def metadata_remove(ctx, did, key):
    """Remove metadata from a DID"""
    args = Arguments({"did": did, "key": key})
    delete_metadata(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@metadata.command("list")
@click.argument("dids", nargs=-1)
@click.option("--plugin", help="Filter down to metadata from specific metadata plugin")
@click.pass_context
def metadata_list_(ctx, dids, plugin):
    """List metadata for a list of DIDs"""
    args = Arguments({"dids": dids, "plugin": plugin})
    get_metadata(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import hashlib
from typing import TYPE_CHECKING, Optional

from sqlalchemy import and_, delete, func, select, update

from rucio.common.exception import DatabaseException
from rucio.common.utils import pid_exists
from rucio.db.sqla.models import Heartbeat
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from threading import Thread
    from typing import TypedDict

    from sqlalchemy.orm import Session

    class HeartbeatDict(TypedDict):
        readable: Optional[str]
        hostname: str
        pid: int
        thread_name: Optional[str]
        updated_at: datetime.datetime
        created_at: datetime.datetime
        payload: Optional[str]
        test: int

DEFAULT_EXPIRATION_DELAY = datetime.timedelta(days=1).total_seconds()


@transactional_session
def sanity_check(
    executable: str,
    hostname: str,
    hash_executable: Optional[str] = None,
    pid: Optional[int] = None,
    thread: Optional["Thread"] = None,
    expiration_delay: float = DEFAULT_EXPIRATION_DELAY,
    *,
    session: "Session"
) -> None:
    """
    sanity_check wrapper to ignore DatabaseException errors.

    :param executable: Executable name as a string, e.g., conveyor-submitter.
    :param hostname: Hostname as a string, e.g., rucio-daemon-prod-01.cern.ch.
    :param hash_executable: Hash of the executable.
    :param pid: UNIX Process ID as a number, e.g., 1234.
    :param thread: Python Thread Object.
    :param expiration_delay: time (in seconds) after which any inactive health check will be removed
    :param session: The database session in use.
    """
    try:
        _sanity_check(executable=executable, hostname=hostname, hash_executable=hash_executable,
                      expiration_delay=expiration_delay, session=session)
        if pid:
            live(executable=executable, hostname=hostname,
                 pid=pid, thread=thread, session=session)
    except DatabaseException:
        pass


@transactional_session
def _sanity_check(
    executable: str,
    hostname: str,
    hash_executable: Optional[str] = None,
    expiration_delay: float = DEFAULT_EXPIRATION_DELAY,
    *,
    session: "Session"
) -> None:
    """
    Check if processes on the host are still running.

    :param executable: Executable name as a string, e.g., conveyor-submitter.
    :param hostname: Hostname as a string, e.g., rucio-daemon-prod-01.cern.ch.
    :param hash_executable: Hash of the executable.
    :param expiration_delay: time (in seconds) after which any inactive health check will be removed
    :param session: The database session in use.
    """
    base_stmt = select(
        Heartbeat.pid
    ).distinct(
    ).where(
        Heartbeat.hostname == hostname
    )
    if executable:
        if not hash_executable:
            hash_executable = calc_hash(executable)

        stmt = base_stmt.where(
            Heartbeat.executable == hash_executable
        )
        for pid in session.execute(stmt).scalars().all():
            if not pid_exists(pid):
                stmt = delete(
                    Heartbeat
                ).where(
                    and_(Heartbeat.executable == hash_executable,
                         Heartbeat.hostname == hostname,
                         Heartbeat.pid == pid)
                )
                session.execute(stmt)
    else:
        for pid in session.execute(base_stmt).scalars().all():
            if not pid_exists(pid):
                stmt = delete(
                    Heartbeat
                ).where(
                    and_(Heartbeat.hostname == hostname,
                         Heartbeat.pid == pid)
                )
                session.execute(stmt)

    if expiration_delay:
        cardiac_arrest(older_than=expiration_delay, session=session)


@transactional_session
def live(
    executable: str,
    hostname: str,
    pid: int,
    thread: Optional["Thread"] = None,
    older_than: int = 600,
    hash_executable: Optional[str] = None,
    payload: Optional[str] = None,
    *,
    session: "Session"
) -> dict[str, int]:
    """
    Register a heartbeat for a process/thread on a given node.
    The executable name is used for the calculation of thread assignments.
    Removal of stale heartbeats is done as a scheduled database job.

    TODO: Returns an assignment dictionary for the given executable.

    :param executable: Executable name as a string, e.g., conveyor-submitter.
    :param hostname: Hostname as a string, e.g., rucio-daemon-prod-01.cern.ch.
    :param pid: UNIX Process ID as a number, e.g., 1234.
    :param thread: Python Thread Object.
    :param older_than: Ignore specified heartbeats older than specified nr of seconds.
    :param hash_executable: Hash of the executable.
    :param payload: Payload identifier which can be further used to identify the work a certain thread is executing.
    :param session: The database session in use.

    :returns heartbeats: Dictionary {assign_thread, nr_threads}
    """
    if not hash_executable:
        hash_executable = calc_hash(executable)

    if thread:
        thread_id = thread.ident
        thread_name = thread.name
    else:
        thread_id = 0
        thread_name = "thread"

    # upsert the heartbeat
    stmt = update(
        Heartbeat
    ).where(
        and_(Heartbeat.executable == hash_executable,
             Heartbeat.hostname == hostname,
             Heartbeat.pid == pid,
             Heartbeat.thread_id == thread_id)
    ).values({
        Heartbeat.updated_at: datetime.datetime.utcnow(),
        Heartbeat.payload: payload
    })
    if not session.execute(stmt).rowcount:
        Heartbeat(executable=hash_executable,
                  readable=executable[:Heartbeat.readable.property.columns[0].type.length],
                  hostname=hostname,
                  pid=pid,
                  thread_id=thread_id,
                  thread_name=thread_name,
                  payload=payload).save(session=session)

    # assign thread identifier
    stmt = select(
        Heartbeat.hostname,
        Heartbeat.pid,
        Heartbeat.thread_id
    ).with_hint(
        Heartbeat,
        'INDEX(HEARTBEATS HEARTBEATS_PK)',
        'oracle'
    ).where(
        and_(Heartbeat.executable == hash_executable,
             Heartbeat.updated_at >= datetime.datetime.utcnow() - datetime.timedelta(seconds=older_than))
    ).group_by(
        Heartbeat.hostname,
        Heartbeat.pid,
        Heartbeat.thread_id
    ).order_by(
        Heartbeat.hostname,
        Heartbeat.pid,
        Heartbeat.thread_id
    )
    result = session.execute(stmt).all()

    # there is no universally applicable rownumber in SQLAlchemy
    # so we have to do it in Python
    assign_thread = 0
    for r in range(len(result)):
        if result[r][0] == hostname and result[r][1] == pid and result[r][2] == thread_id:
            assign_thread = r
            break

    return {'assign_thread': assign_thread,
            'nr_threads': len(result)}


@transactional_session
def die(
    executable: str,
    hostname: str,
    pid: int,
    thread: "Thread",
    older_than: Optional[int] = None,
    hash_executable: Optional[str] = None,
    *,
    session: "Session"
) -> None:
    """
    Remove a single heartbeat older than specified.

    :param executable: Executable name as a string, e.g., conveyor-submitter
    :param hostname: Hostname as a string, e.g., rucio-daemon-prod-01.cern.ch
    :param pid: UNIX Process ID as a number, e.g., 1234
    :param thread: Python Thread Object
    :param older_than: Removes specified heartbeats older than specified nr of seconds
    :param hash_executable: Hash of the executable.
    :param session: The database session in use.
    """
    if not hash_executable:
        hash_executable = calc_hash(executable)

    stmt = delete(
        Heartbeat
    ).where(
        and_(Heartbeat.executable == hash_executable,
             Heartbeat.hostname == hostname,
             Heartbeat.pid == pid,
             Heartbeat.thread_id == thread.ident)
    )

    if older_than:
        stmt = stmt.where(
            Heartbeat.updated_at < datetime.datetime.utcnow() - datetime.timedelta(seconds=older_than)
        )
    session.execute(stmt)


@transactional_session
def cardiac_arrest(older_than: Optional[int] = None, *, session: "Session") -> None:
    """
    Removes all heartbeats older than specified.

    :param older_than: Removes all heartbeats older than specified nr of seconds
    :param session: The database session in use.
    """

    stmt = delete(
        Heartbeat
    )

    if older_than:
        stmt = stmt.where(
            Heartbeat.updated_at < datetime.datetime.utcnow() - datetime.timedelta(seconds=older_than)
        )
    session.execute(stmt)


@read_session
def list_heartbeats(*, session: "Session") -> list["HeartbeatDict"]:
    """
    List all heartbeats.

    :param session: The database session in use.

    :returns: List of dicts
    """

    stmt = select(
        Heartbeat.readable,
        Heartbeat.hostname,
        Heartbeat.pid,
        Heartbeat.thread_name,
        Heartbeat.updated_at,
        Heartbeat.created_at,
        Heartbeat.payload
    ).order_by(
        Heartbeat.readable,
        Heartbeat.hostname,
        Heartbeat.thread_name
    )

    result = session.execute(stmt).all()
    json_result = []
    for element in result:
        json_result.append(element._asdict())
    return json_result


@read_session
def list_payload_counts(
    executable: str,
    older_than: int = 600,
    hash_executable: Optional[str] = None,
    *,
    session: "Session"
) -> dict[str, int]:
    """
    Give the counts of number of threads per payload for a certain executable.

    :param executable: Executable name as a string, e.g., conveyor-submitter
    :param older_than: Removes specified heartbeats older than specified nr of seconds
    :param hash_executable: Hash of the executable.
    :param session: The database session in use.

    :returns: Dict
    """

    if not hash_executable:
        hash_executable = calc_hash(executable)
    stmt = select(
        Heartbeat.payload,
        func.count(Heartbeat.payload)
    ).where(
        and_(Heartbeat.executable == hash_executable,
             Heartbeat.updated_at >= datetime.datetime.utcnow() - datetime.timedelta(seconds=older_than))
    ).group_by(
        Heartbeat.payload
    ).order_by(
        Heartbeat.payload
    )

    return dict((payload, count) for payload, count in session.execute(stmt).all() if payload)


def calc_hash(executable: str) -> str:
    """
    Calculates a SHA256 hash.

    return: String of hexadecimal hash
    """
    return hashlib.sha256(executable.encode('utf-8')).hexdigest()



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio_admin import add_scope, list_scopes
from rucio.cli.utils import Arguments


@click.group()
def scope():
    """Interact with scopes - a logical grouping of DIDs"""


@scope.command("add")
@click.argument("scope-name")
@click.option("-a", "--account", help="Associated account", required=True)
@click.pass_context
def add_(ctx, account, scope_name):
    """Add a new scope with name [SCOPE-NAME]"""
    args = Arguments({"scope": scope_name, "account": account})
    add_scope(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@scope.command("list")
@click.option("-a", "--account", help="Filter by associated account", required=False)
@click.pass_context
def list_(ctx, account):
    """List existing scopes"""
    list_scopes(Arguments({"account": account}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utility functions for stomp connections
"""
import json
import logging
import random
import socket
from collections import namedtuple
from copy import deepcopy
from functools import partial
from time import monotonic
from typing import TYPE_CHECKING, Any

from stomp import Connection12
from stomp.exception import ConnectFailedException, NotConnectedException
from stomp.listener import HeartbeatListener

from rucio.common.config import config_get, config_get_bool, config_get_float, config_get_int, config_get_list
from rucio.common.logging import formatted_logger
from rucio.core.monitor import MetricManager

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from stomp.connect import Frame

    from rucio.common.types import LoggerFunction


METRICS = MetricManager(module=__name__)


class Connection(Connection12):
    """
    Connection class.

    Wraps Stomp Connection but knows the brokers without accessing
    hidden variables from the Transport.
    """
    def __init__(self, host_and_ports: list[tuple[str, int]], **kwargs):
        """
        Initialise.

        Args:
            host_and_ports: brokers list

        Kwargs:
            Arguments to pass to the Constructor12 base class.
        """
        super().__init__(host_and_ports=host_and_ports, **kwargs)
        self._brokers = host_and_ports

    @property
    def brokers(self) -> list[tuple[str, int]]:
        """
        List brokers.

        Returns:
            All assigned brokers in (host, port) format.
        """
        return self._brokers


class ListenerBase(HeartbeatListener):
    """Listener Base."""

    _logger = formatted_logger(logging.log, 'ListenerBase %s')

    def __init__(self,
                 conn: Connection,
                 logger: "None | LoggerFunction" = None,
                 **kwargs):
        """
        Initialise.

        Args:
            conn: The connection object that is using this listener
            logger: Logger to use. Defaults to logging.getLogger(__name__).getChild(__qualname__).

        Kwargs:
            Arguments to pass to the stomp.ConnectionListener base class.
        """
        super().__init__(transport=conn.transport, **kwargs)
        self._conn = conn
        if logger is not None:
            self._logger = logger

    @METRICS.count_it
    def on_heartbeat_timeout(self):
        self._conn.disconnect()

    @METRICS.count_it
    def on_error(self, frame: "Frame"):
        """
        on_error
        """
        self._logger(logging.ERROR, 'Message receive error: [%s] %s', self._conn.brokers[0][0], frame.body)


StompConfig = namedtuple("StompConfig", ('brokers', 'use_ssl', 'port', 'vhost',
                                         'destination', 'key_file', 'cert_file',
                                         'username', 'password', 'nonssl_port',
                                         'reconnect_attempts_max', 'timeout', 'heartbeats'))


class StompConnectionManager:
    """Stomp Connection Manager."""

    _logger = formatted_logger(logging.log, 'StompConnectionManager %s')

    def __init__(self,
                 config_section: str,
                 logger: "None | LoggerFunction" = None):
        """
        Initialise.

        Args:
            config_section: The name of the config section for this manager to parse for configuration.
            logger: logger to use. Defaults to logging.getLogger(__name__).getChild(__qualname__).
        """
        if logger is not None:
            self._logger = logger
        self._config = self._parse_config(config_section)
        self._listener_factory = None
        self._conns = []
        for broker in self._config.brokers:
            conn = Connection(host_and_ports=[broker],
                              vhost=self._config.vhost,
                              reconnect_attempts_max=self._config.reconnect_attempts_max,
                              timeout=self._config.timeout,
                              heartbeats=self._config.heartbeats)
            if self._config.use_ssl:
                conn.set_ssl(cert_file=self._config.cert_file, key_file=self._config.key_file)
            self._conns.append(conn)

    @property
    def config(self) -> StompConfig:
        """
        Get the config.

        Returns:
            config object.
        """
        return deepcopy(self._config)

    def set_listener_factory(self, name: str, listener_cls: type, **kwargs) -> None:
        """
        Setup listener factory

        This method will setup a factory to create a name and listener for the arguments to
        connection.set_listener based on pre-defined argument values.

        Args:
            name: Listener name
            listener_cls: Listener class.
        """
        def create_listener(name, listener_factory, conn):
            return name, listener_factory(conn=conn)
        self._listener_factory = partial(create_listener,
                                         name=name,
                                         listener_factory=partial(listener_cls, logger=self._logger, **kwargs))

    def _parse_config(self, config_section: str) -> StompConfig:
        """
        Parse config section.

        Args:
            config_section: The name of the config section for this manager to parse for configuration.

        Raises:
            RuntimeError: If cannot parse config sections 'brokers' or 'use_ssl' or if misconfigured.

        Returns:
            Stomp manager configuration object.
        """
        try:
            brokers = config_get(config_section, 'brokers')
        except Exception as exc:
            self._logger(logging.ERROR, "Could not load brokers from configuration")
            raise RuntimeError('Could not load brokers from configuration') from exc

        try:
            use_ssl = config_get_bool(config_section, 'use_ssl')
        except Exception as exc:
            self._logger(logging.ERROR, "could not find use_ssl in configuration -- please update your rucio.cfg")
            raise RuntimeError('could not find use_ssl in configuration -- please update your rucio.cfg') from exc

        port = config_get_int(config_section, 'port')
        vhost = config_get(config_section, 'broker_virtual_host', raise_exception=False)
        destination = config_get(config_section, "destination")
        key_file = config_get(config_section, 'ssl_key_file', default=None, raise_exception=False)
        cert_file = config_get(config_section, 'ssl_cert_file', default=None, raise_exception=False)
        username = config_get(config_section, 'username', default=None, raise_exception=False)
        password = config_get(config_section, 'password', default=None, raise_exception=False)
        nonssl_port = config_get_int(config_section, 'nonssl_port', default=0, raise_exception=False)
        timeout = config_get_float(config_section, 'timeout', default=None, raise_exception=False)
        heartbeats = tuple(config_get_list(config_section, 'heartbeats', default=[0., 0.], raise_exception=False))
        reconnect_attempts = config_get_int(config_section, 'reconnect_attempts', default=100)
        if use_ssl and (key_file is None or cert_file is None):
            self._logger(logging.ERROR, "If use_ssl is True in config you must provide both 'ssl_cert_file' "
                                        "and 'ssl_key_file'")
            raise RuntimeError("If use_ssl is True in config you must provide both 'ssl_cert_file' and 'ssl_key_file'")
        if not use_ssl and (username is None or password is None or nonssl_port == 0):
            self._logger(logging.ERROR, "If use_ssl is False in config you must provide "
                                        "'username', 'password' and 'nonssl_port'")
            raise RuntimeError("If use_ssl is False in config you must provide "
                               "'username', 'password' and 'nonssl_port'")
        return StompConfig(brokers=self._resolve_host_and_port(brokers, port if use_ssl else nonssl_port),
                           use_ssl=use_ssl,
                           port=port, vhost=vhost,
                           destination=destination, key_file=key_file, cert_file=cert_file,
                           username=username, password=password, nonssl_port=nonssl_port,
                           reconnect_attempts_max=reconnect_attempts, timeout=timeout, heartbeats=heartbeats)

    def _resolve_host_and_port(self, fqdns: "str | Iterable[str]", port: int) -> list[tuple[str, int]]:
        """
        Resolve host and port.

        Args:
            fqdns: fully qualified domain name(s)
            port: port

        Returns:
            list of (host, port) tuples.
        """
        if isinstance(fqdns, str):
            fqdns = fqdns.split(',')

        hosts_and_ports = []
        for fqdn in fqdns:
            try:
                addrinfos = socket.getaddrinfo(fqdn.strip(), port, socket.AF_INET, 0, socket.IPPROTO_TCP)
            except socket.gaierror as exc:
                self._logger(logging.ERROR, "[broker] Cannot resolve domain name %s (%s)", fqdn.strip(), str(exc))
                continue

            hosts_and_ports.extend(addrinfo[4] for addrinfo in addrinfos)
        if not hosts_and_ports:
            self._logger(logging.WARNING, "[broker] No resolved brokers")
        return hosts_and_ports

    def _is_stalled(self, conn: Connection) -> bool:
        """
        Determine if a connection is stalled.

        Args:
            conn: The Connection object

        Returns:
            Whether the connection has stalled.
        """
        received_heartbeat = getattr(conn, 'received_heartbeat', None)
        if received_heartbeat is None or not any(self._config.heartbeats):
            return False

        heartbeat_period_seconds = max(0, self._config.heartbeats[0], self._config.heartbeats[1]) / 1000
        if heartbeat_period_seconds == 0.:
            return False

        now = monotonic()
        if received_heartbeat + 10 * heartbeat_period_seconds >= now:
            return False

        return True

    def connect(self) -> "Iterator[Connection]":
        """
        Connect.

        Yields:
            Each connection object after ensuring it's connected.
        """
        config = self._config
        params = {'wait': True, "heartbeats": self._config.heartbeats}
        self._logger(logging.WARNING, 'heartbeats: %s', self._config.heartbeats)
        if config.use_ssl:
            params.update(username=config.username, password=config.password)

        for conn in self._conns:
            if self._is_stalled(conn):
                try:
                    conn.disconnect()
                except Exception:
                    self._logger(logging.ERROR, "[broker] Stalled connection could not be disconnected")
            if not conn.is_connected():
                self._logger(logging.INFO, 'connecting to %s:%s', *conn.brokers[0])
                METRICS.counter('reconnect.{host}').labels(host=conn.brokers[0][0]).inc()
                if self._listener_factory is not None:
                    conn.set_listener(*self._listener_factory(conn=conn))

                try:
                    conn.connect(**params)
                except ConnectFailedException as error:
                    self._logger(logging.WARNING, "[broker] Could not deliver message due to "
                                                  "ConnectFailedException: %s", str(error))
                    continue
                except Exception as error:
                    self._logger(logging.ERROR, "[broker] Could not connect: %s", str(error))
                    continue
            try:
                yield conn
            except Exception:
                self._logger(logging.ERROR, "[broker] Error in yielded code, skipping to next connection.")

    def deliver_messages(self, messages: "Iterable[dict[str, Any]]") -> list[int]:
        """
        Deliver messages.

        Args:
            messages: Messages to deliver.

        Returns:
            delivered message ids, ready for deletion.
        """
        config = self._config
        conn = random.sample(list(self.connect()), 1)[0]
        to_delete = []
        for message in messages:
            try:
                body = json.dumps({"event_type": str(message["event_type"]).lower(),
                                   "payload": message["payload"],
                                   "created_at": str(message["created_at"])})
            except ValueError:
                self._logger(logging.ERROR, "[broker] Cannot serialize payload to JSON: %s", str(message["payload"]))
                to_delete.append(message["id"])
                continue

            try:
                conn.send(
                    body=body,
                    destination=config.destination,
                    headers={"persistent": "true",
                             "event_type": str(message["event_type"]).lower()}
                )
                to_delete.append(message["id"])
            except NotConnectedException as error:
                self._logger(logging.WARNING, "[broker] Could not deliver message due to NotConnectedException: %s",
                             str(error))
                continue
            except Exception as error:
                self._logger(logging.ERROR, "[broker] Could not deliver message: %s", str(error))
                continue

            msg_event_type = str(message["event_type"]).lower()
            msg_payload = message.get("payload", {})
            if msg_event_type.startswith("transfer") or msg_event_type.startswith("stagein"):
                self._logger(logging.DEBUG,
                             "[broker] - event_type: %s, scope: %s, name: %s, rse: %s, request-id: %s, "
                             "transfer-id: %s, created_at: %s",
                             msg_event_type,
                             msg_payload.get("scope", None),
                             msg_payload.get("name", None),
                             msg_payload.get("dst-rse", None),
                             msg_payload.get("request-id", None),
                             msg_payload.get("transfer-id", None),
                             str(message["created_at"]))

            elif msg_event_type.startswith("dataset"):
                self._logger(logging.DEBUG,
                             "[broker] - event_type: %s, scope: %s, name: %s, rse: %s, rule-id: %s, created_at: %s)",
                             msg_event_type,
                             msg_payload.get("scope", None),
                             msg_payload.get("name", None),
                             msg_payload.get("rse", None),
                             msg_payload.get("rule_id", None),
                             str(message["created_at"]))

            elif msg_event_type.startswith("deletion"):
                if "url" not in msg_payload:
                    msg_payload["url"] = "unknown"
                self._logger(logging.DEBUG,
                             "[broker] - event_type: %s, scope: %s, name: %s, rse: %s, url: %s, created_at: %s)",
                             msg_event_type,
                             msg_payload.get("scope", None),
                             msg_payload.get("name", None),
                             msg_payload.get("rse", None),
                             msg_payload.get("url", None),
                             str(message["created_at"]))
            else:
                self._logger(logging.DEBUG, "[broker] Other message: %s", message)

        return to_delete

    def subscribe(self, id_: str, ack: str, destination: "None | str" = None, **kwargs) -> None:
        """
        Subscribe

        Args:
            id_: The identifier to uniquely identify the subscription
            ack: Either auto, client or client-individual
            destination: The topic or queue to subscribe to. If None then
                         destination is taken from the rucio config Defaults to None.

        Kwargs:
            Arguments to pass to the Construction objects subscribe method.
        """
        if destination is None:
            destination = self._config.destination
        for conn in self.connect():
            conn.subscribe(destination=destination,
                           id=id_, ack=ack, **kwargs)

    def disconnect(self) -> None:
        """Disconnect."""
        for conn in self._conns:
            try:
                conn.disconnect()
            except Exception:
                self._logger(logging.ERROR, "[broker] Could not disconnect")



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pydoc
import re
import shutil
import sys
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from rich import box
from rich.console import Console, JustifyMethod, RenderableType
from rich.logging import RichHandler
from rich.table import Table
from rich.text import Text

from rucio.common.config import config_get

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from rich.style import StyleType


MIN_CONSOLE_WIDTH = 80
MAX_TRACEBACK_WIDTH = 120  # Slightly higher than default width of rich.traceback (100).


class CLITheme:
    """
    Class to define styles for Rich widgets and prints in the CLI.
    """
    TABLE_FMT = box.SQUARE

    TEXT_HIGHLIGHT = 'grey50'  # Used to highlight prints between tables, e.g. get_metadata or stat.
    SUBHEADER_HIGHLIGHT = 'cyan'  # Used to highlight prints between tables for a section, e.g. protocols or usage per account.
    SPINNER = 'dots2'
    SPINNER_STYLE = 'green'

    JSON_STR = 'green'
    JSON_NUM = 'bold cyan'

    SUCCESS_ICON = '[bold green]\u2714[/]'
    FAILURE_ICON = '[bold red]\u2717[/]'

    LOG_THEMES = {
        'logging.level.info': 'default',
        'logging.level.warning': 'bold yellow',
        'logging.level.debug': 'dim',
        'logging.level.critical': 'default on red',
        'repr.bool_true': 'green',
        'repr.bool_false': 'red',
        'log.time': 'turquoise4'
    }

    DID_TYPE = {
        'CONTAINER': 'bold dodger_blue1',
        'DATASET': 'bold orange3',
        'FILE': 'bold default',
        'COLLECTION': 'bold green',
        'ALL': 'bold magenta',
        'DERIVED': 'bold magenta'
    }

    BOOLEAN = {
        'True': 'green',
        'False': 'red'
    }

    RULE_STATE = {
        'OK': 'bold green',
        'REPLICATING': 'bold default',
        'STUCK': 'bold orange3',
        'WAITING APPROVAL': 'bold dodger_blue1',
        'INJECT': 'bold medium_purple3',
        'SUSPENDED': 'bold red'
    }

    RSE_TYPE = {
        'DISK': 'bold dodger_blue1',
        'TAPE': 'bold orange3',
        'UNKNOWN': 'bold'
    }

    SUBSCRIPTION_STATE = {
        'ACTIVE': 'bold green',
        'INACTIVE': 'bold',
        'NEW': 'bold dodger_blue1',
        'UPDATED': 'bold green',
        'BROKEN': 'bold red',
        'UNKNOWN': 'bold orange3'
    }

    AVAILABILITY = {
        'AVAILABLE': 'bold green',
        'DELETED': 'bold',
        'LOST': 'bold red',
        'UNKNOWN': 'bold orange3'
    }

    ACCOUNT_STATUS = {
        'ACTIVE': 'bold green',
        'SUSPENDED': 'bold red',
        'DELETED': 'bold'
    }

    REPLICA_STATE = {
        'A': 'bold green',
        'U': 'bold red',
        'C': 'bold default',
        'B': 'bold dodger_blue1',
        'D': 'bold red',
        'T': 'bold orange3'
    }

    ACCOUNT_TYPE = {
        'USER': 'default',
        'GROUP': 'medium_purple3',
        'SERVICE': 'yellow'
    }


def setup_rich_logger(
    module_name: Optional[str] = None,
    logger_name: Optional[str] = None,
    logger_level: Optional[int] = None,
    verbose: bool = False,
    console: Optional[Console] = None
) -> logging.Logger:
    """
    Factory method to set logger with RichHandler.

    The function is a copy of the method in rucio.common.utils setup_logger() with minor changes.

    :param module_name: __name__ of the module that is calling this method
    :param logger_name: name of the logger, typically name of the module.
    :param logger_level: if not given, fetched from config.
    :param verbose: verbose option set in bin/rucio
    :param console: Rich console object
    :returns: logger with RichHandler
    """
    # Helper method for cfg check.
    def _force_cfg_log_level(cfg_option: str) -> bool:
        cfg_forced_modules = config_get('logging', cfg_option, raise_exception=False, default=None, clean_cached=True, check_config_table=False)
        if cfg_forced_modules and module_name is not None:
            if re.match(str(cfg_forced_modules), module_name):
                return True
        return False

    if not logger_name:
        if not module_name:
            logger_name = 'user'
        else:
            logger_name = module_name.split('.')[-1]
    logger = logging.getLogger(logger_name)

    # Extracting the log level.
    if not logger_level:
        logger_level = logging.INFO
        if verbose:
            logger_level = logging.DEBUG

        # Overriding by the config.
        cfg_levels = (logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR)
        for level in cfg_levels:
            cfg_opt = 'forceloglevel' + logging.getLevelName(level)
            if _force_cfg_log_level(cfg_opt):
                logger_level = level

    logger.setLevel(logger_level)

    def add_handler(logger: logging.Logger,
                    console: Optional[Console] = None,
                    verbose: bool = False) -> None:

        def time_formatter(timestamp: datetime) -> Text:
            return Text(f"[{timestamp.isoformat(sep=' ', timespec='milliseconds')}]")

        console = console or Console()
        handler = RichHandler(rich_tracebacks=True, markup=True, show_path=verbose, show_time=verbose, console=console, tracebacks_width=min(console.width, MAX_TRACEBACK_WIDTH),
                              tracebacks_word_wrap=True, log_time_format=time_formatter)
        logger.addHandler(handler)

    # Setting handler and formatter.
    if not logger.handlers:
        add_handler(logger, console, verbose)

    return logger


def _format_value(value: Optional[Union[RenderableType, int, float, bool, datetime]] = None) -> RenderableType:
    """
    Formats the value based on its type for Rich Table.

    A helper function to format the value to Rich RenderableType.

    :param value: value to format
    :returns: formatted value
    """
    if value is None or str(value) == 'None':
        return ''
    if isinstance(value, bool):
        return CLITheme.BOOLEAN[str(value)]
    if isinstance(value, (int, float, datetime)):
        return str(value)
    return value


def generate_table(
    rows: 'Sequence[Sequence[Union[RenderableType, int, float, bool, datetime]]]',
    headers: Optional['Sequence[RenderableType]'] = None,
    row_styles: Optional['Sequence[StyleType]'] = None,
    col_alignments: Optional[list[JustifyMethod]] = None,
    table_format: box.Box = CLITheme.TABLE_FMT,
) -> Table:
    """
    Generates a Rich Table object from given input rows.

    The elements in each row can be either plain strings or Rich renderable objects.
    Passing strings will display them as simple text, while using Rich objects
    allows you to introduce additional structure, styling, and widgets (e.g. Text, Trees) into
    the table. Strings with style markup will be rendered as styled text.

    :param table_format: style of the table
    :param headers: list of headers
    :param rows: list of rows
    :param col_alignments: list of column alignments
    :param row_styles: list of row styles
    :returns: a Rich Table object
    """
    table = Table(box=table_format, show_header=headers is not None and len(headers) > 0)
    table.row_styles = row_styles or ['none', 'dim']

    if len(rows) == 0:
        if headers:
            for header in headers:
                table.add_column(header)
        return table

    # Auto-detect on first row, numerical values on the right.
    col_alignments = col_alignments or ['right' if str(col).isnumeric() else 'left' for col in rows[0]]
    headers = headers or [''] * len(rows[0])
    while len(headers) > len(col_alignments):
        col_alignments.append('left')

    for header, alignment in zip(headers, col_alignments):
        table.add_column(header, overflow='fold', justify=alignment)

    for row in rows:
        row = [_format_value(col) for col in row]
        table.add_row(*row)
    return table


def print_output(
    *output: Any,
    console: Console,
    no_pager: bool = False
) -> None:
    """
    Prints the objects using the specified Rich console object. Optionally disables the pager if specified.

    The function works similarly to Rich's `console.print()` method but provides additional control over the pager feature.

    :param output: objects to print to the terminal
    :param console: Rich console object
    :param no_pager: flag to disable the pager
    """
    if console.is_terminal:
        if no_pager:
            console.print(*output)
        else:
            console.width = sys.maxsize   # Overwrite auto-detected console width.
            console.begin_capture()
            console.print(*output)
    else:
        console.width = sys.maxsize
        console.print(*output)


def get_cli_config() -> str:
    """
    Returns the CLI type from the config file.

    :returns: CLI type (Rich or tabulate)
    """
    cli_type = config_get('experimental', 'cli', raise_exception=False, default='tabulate').lower()
    if cli_type not in ['rich', 'tabulate']:
        cli_type = 'tabulate'
    return cli_type


def get_pager() -> 'Callable[[str], None]':
    """
    Returns the pager function based on the terminal availability.

    :returns: pager
    """
    default_pager = 'less'
    # Attempt to use the default pager if available.
    if shutil.which(default_pager) is not None:
        return lambda text: pydoc.pipepager(text, f'{default_pager} -FRSXKM')

    # Fall back to pydoc.pager if the default pager is not available.
    return pydoc.pager



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import socket
import threading
import time
from typing import TYPE_CHECKING

import rucio.db.sqla.util
from rucio.common import exception
from rucio.common.logging import setup_logging
from rucio.common.utils import get_thread_with_periodic_running_function
from rucio.core.did import create_reports
from rucio.core.heartbeat import die, live, sanity_check

if TYPE_CHECKING:
    from types import FrameType
    from typing import Optional

graceful_stop = threading.Event()
DAEMON_NAME = 'rucio-follower'


def aggregate_events(
        once: bool = False
) -> None:
    """
    Collect all the events affecting the dids followed by the corresponding account.
    """

    logging.info('event_aggregation: started')

    hostname = socket.gethostname()
    pid = os.getpid()
    current_thread = threading.current_thread()
    live(executable=DAEMON_NAME, hostname=hostname, pid=pid, thread=current_thread)

    while not graceful_stop.is_set():
        heartbeat = live(executable=DAEMON_NAME, hostname=hostname, pid=pid, thread=current_thread)
        # Create a report of events and send a mail to the corresponding account.
        start_time = time.time()
        create_reports(total_workers=heartbeat['nr_threads'] - 1,
                       worker_number=heartbeat['assign_thread'])
        logging.info('worker[%s/%s] took %s for creating reports' % (heartbeat['assign_thread'], heartbeat['nr_threads'] - 1, time.time() - start_time))

        if once:
            break

    logging.info('follower: graceful stop requested')
    die(executable=DAEMON_NAME, hostname=hostname, pid=pid, thread=current_thread)
    logging.info('follower: graceful stop done')


def stop(signum: "Optional[int]" = None, frame: "Optional[FrameType]" = None) -> None:
    """
    Graceful exit.
    """
    graceful_stop.set()


def run(
        once: bool = False,
        threads: int = 1
) -> None:
    """
    Starts up the follower threads
    """
    setup_logging(process_name=DAEMON_NAME)

    if rucio.db.sqla.util.is_old_db():
        raise exception.DatabaseException('Database was not updated, daemon won\'t start')

    hostname = socket.gethostname()
    sanity_check(executable=DAEMON_NAME, hostname=hostname)

    if once:
        logging.info("executing one follower iteration only")
        aggregate_events(once)
    else:
        logging.info("starting follower threads")
        # Run the follower daemon thrice a day
        thread_list = [get_thread_with_periodic_running_function(28800, aggregate_events, graceful_stop) for i in range(threads)]
        [t.start() for t in thread_list]

        logging.info("waiting for interrupts")
        # Interruptible joins require a timeout.
        while thread_list[0].is_alive():
            [t.join(timeout=3.14) for t in thread_list]



#!/usr/bin/env python3
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from datetime import datetime, timedelta
from os import remove
from time import sleep

import pytest

from rucio.core import rse_expression_parser
from rucio.core.did import set_metadata
from rucio.core.replica import list_bad_replicas_status, list_replicas, update_replica_state
from rucio.core.rse import add_rse_attribute
from rucio.daemons.replicarecoverer.suspicious_replica_recoverer import run, stop
from rucio.db.sqla.constants import BadFilesStatus, DIDType, ReplicaState
from rucio.tests.common import execute


class TestReplicaRecoverer:

    @pytest.fixture(autouse=True)
    def setup_obj(self, vo, rse_factory, replica_client, mock_scope, file_factory, scope_factory):

        # Using two test RSEs
        self.rse4suspicious, self.rse4suspicious_id = rse_factory.make_posix_rse(deterministic=True, vo=vo)
        self.rse4recovery, self.rse4recovery_id = rse_factory.make_posix_rse(deterministic=True, vo=vo)

        # Create new scopes
        _, [self.scope_declarebad] = scope_factory(vos=[vo])
        _, [self.scope_nopolicy] = scope_factory(vos=[vo])
        _, [self.scope_ignore] = scope_factory(vos=[vo])

        # For testing, we create 5 files and upload them to Rucio to two test RSEs.
        self.tmp_file1 = file_factory.file_generator()
        self.tmp_file2 = file_factory.file_generator()
        self.tmp_file3 = file_factory.file_generator()
        self.tmp_file4 = file_factory.file_generator()
        self.tmp_file5 = file_factory.file_generator()
        self.tmp_file6 = file_factory.file_generator()
        self.tmp_file7 = file_factory.file_generator()
        self.tmp_file8 = file_factory.file_generator()
        self.tmp_file9 = file_factory.file_generator()
        self.tmp_file10 = file_factory.file_generator()
        self.tmp_file11 = file_factory.file_generator()  # tmp_file11 shouldn't be declare as bad, as it doesn't have a data type.
        self.tmp_file12 = file_factory.file_generator()  # tmp_file12 is used to test the creation of rules by the daemon.
        self.tmp_file13 = file_factory.file_generator()

        self.listdids_mock = [{'scope': mock_scope, 'name': f.name, 'type': DIDType.FILE}
                              for f in [self.tmp_file1, self.tmp_file2, self.tmp_file3, self.tmp_file4, self.tmp_file5, self.tmp_file6, self.tmp_file12]]
        self.listdids_declarebad = [{'scope': self.scope_declarebad, 'name': f.name, 'type': DIDType.FILE}
                                    for f in [self.tmp_file7, self.tmp_file9, self.tmp_file11, self.tmp_file13]]
        self.listdids_nopolicy = [{'scope': self.scope_nopolicy, 'name': f.name, 'type': DIDType.FILE}
                                  for f in [self.tmp_file8]]
        self.listdids_ignore = [{'scope': self.scope_ignore, 'name': f.name, 'type': DIDType.FILE}
                                for f in [self.tmp_file10]]

        for rse in [self.rse4suspicious, self.rse4recovery]:
            # Upload files with scope "mock_scope"
            cmd = 'rucio -v upload --legacy --rse {0} --scope {1} {2} {3} {4} {5} {6} {7} {8}'.format(rse, mock_scope.external, self.tmp_file1, self.tmp_file2, self.tmp_file3, self.tmp_file4, self.tmp_file5, self.tmp_file6, self.tmp_file12)
            exitcode, out, err = execute(cmd)
            print("mock_scope:", exitcode, out, err)
            # checking if Rucio upload went OK
            assert exitcode == 0

            # Upload files with scope "scope_declarebad"
            cmd = 'rucio -v upload --legacy --rse {0} --scope {1} {2} {3} {4} {5}'.format(rse, self.scope_declarebad.external, self.tmp_file7, self.tmp_file9, self.tmp_file11, self.tmp_file13)
            exitcode, out, err = execute(cmd)
            print("scope_declarebad:", exitcode, out, err)
            # checking if Rucio upload went OK
            assert exitcode == 0

            # Upload files with scope "scope_nopolicy"
            cmd = 'rucio -v upload --legacy --rse {0} --scope {1} {2}'.format(rse, self.scope_nopolicy.external, self.tmp_file8)
            exitcode, out, err = execute(cmd)
            print("scope_nopolicy:", exitcode, out, err)
            # checking if Rucio upload went OK
            assert exitcode == 0

            # Upload files with scope "scope_nopolicy"
            cmd = 'rucio -v upload --legacy --rse {0} --scope {1} {2}'.format(rse, self.scope_ignore.external, self.tmp_file10)
            exitcode, out, err = execute(cmd)
            print("scope_ignore:", exitcode, out, err)
            # checking if Rucio upload went OK
            assert exitcode == 0

        # Explanation of the fictional data types:
        # testtypedeclarebad: Files are specified to be declared bad
        # testtypeignore: Files are specified to be ignored
        # testtypenopolicy: Files either have no policy or no recognised policy and are ignored by default

        # Set fictional datatypes
        set_metadata(mock_scope, self.tmp_file3.name, 'datatype', 'RAW')
        set_metadata(mock_scope, self.tmp_file4.name, 'datatype', 'testtypedeclarebad')
        set_metadata(mock_scope, self.tmp_file5.name, 'datatype', 'testtypenopolicy')
        set_metadata(mock_scope, self.tmp_file6.name, 'datatype', 'testtypeignore')
        set_metadata(self.scope_declarebad, self.tmp_file7.name, 'datatype', 'testtypenopolicy')
        set_metadata(self.scope_nopolicy, self.tmp_file8.name, 'datatype', 'testtypenopolicy')
        set_metadata(self.scope_declarebad, self.tmp_file9.name, 'datatype', 'testtypeignore')
        set_metadata(self.scope_ignore, self.tmp_file10.name, 'datatype', 'testtypedeclarebad')
        set_metadata(self.scope_declarebad, self.tmp_file13.name, 'datatype', 'testtypedryrun')
        # tmp_file1, 2, 11 and 12 don't have a datatypes.

        # Allow for the RSEs to be affected by the suspicious file recovery daemon
        add_rse_attribute(self.rse4suspicious_id, "enable_suspicious_file_recovery", True)
        add_rse_attribute(self.rse4recovery_id, "enable_suspicious_file_recovery", True)

        # removing physical files from /tmp location - keeping only their DB info
        remove(self.tmp_file1)
        remove(self.tmp_file2)
        remove(self.tmp_file3)
        remove(self.tmp_file4)
        remove(self.tmp_file5)
        remove(self.tmp_file6)
        remove(self.tmp_file7)
        remove(self.tmp_file8)
        remove(self.tmp_file9)
        remove(self.tmp_file10)
        remove(self.tmp_file11)
        remove(self.tmp_file12)
        remove(self.tmp_file13)

        # Reset the cache to include the new RSEs
        rse_expression_parser.REGION.invalidate()

        # Gather replica info
        replicalist_scope_mock = list(list_replicas(dids=self.listdids_mock))
        replicalist_scope_declarebad = list(list_replicas(dids=self.listdids_declarebad))
        replicalist_scope_nopolicy = list(list_replicas(dids=self.listdids_nopolicy))
        replicalist_scope_ignore = list(list_replicas(dids=self.listdids_ignore))

        # Changing the replica statuses as follows:
        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # Name         State(s) declared on rse4recovery       State(s) declared on rse4suspicious        Scope                     Metadata "datatype"
        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # tmp_file1    available                                suspicious (available)                    mock_scope                  <none>
        # tmp_file2    available                                suspicious + bad (unavailable)            mock_scope                  <none>
        # tmp_file3    unavailable                              suspicious (available)                    mock_scope                  RAW
        # tmp_file4    unavailable                              suspicious (available)                    mock_scope                  testtypedeclarebad
        # tmp_file5    unavailable                              suspicious (available)                    mock_scope                  testtypenopolicy
        # tmp_file6    unavailable                              suspicious (available)                    mock_scope                  testtypeignore
        # tmp_file7    unavailable                              suspicious (available)                    scope_declarebad            testtypenopolicy
        # tmp_file8    unavailable                              suspicious (available)                    scope_nopolicy              testtypenopolicy
        # tmp_file9    unavailable                              suspicious (available)                    scope_declarebad            testtypeignore
        # tmp_file10   unavailable                              suspicious (available)                    scope_ignore                testtypedeclarebad
        # tmp_file11   unavailable                              suspicious (available)                    scope_declarebad            <none>
        # tmp_file12   unavailable                              suspicious (available)                    mock_scope                  <none>
        # tmp_file13   unavailable                              suspicious (available)                    scope_declarebad            testtypedryrun
        # ----------------------------------------------------------------------------------------------------------------------------------------------------

        for replica in replicalist_scope_mock:
            suspicious_pfns = replica['rses'][self.rse4suspicious_id]
            # Declare each file as suspicious multiple times, apart from tmp_file12, which
            # should only be declared suspicious once. tmp_file12 is used to test the
            # creation of rules by the daemon.
            if replica['name'] == self.tmp_file12.name:
                print("Declaring suspicious file replica: " + suspicious_pfns[0])
                # The reason must contain the word "checksum", so that the replica can be declared bad.
                replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
                sleep(1)
            else:
                for i in range(3):
                    print("Declaring suspicious file replica: " + suspicious_pfns[0])
                    # The reason must contain the word "checksum", so that the replica can be declared bad.
                    replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
                    sleep(1)
            if replica['name'] == self.tmp_file2.name:
                print("Declaring bad file replica: " + suspicious_pfns[0])
                replica_client.declare_bad_file_replicas([suspicious_pfns[0], ], 'checksum')
            if replica['name'] == self.tmp_file3.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, mock_scope, self.tmp_file3.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file4.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, mock_scope, self.tmp_file4.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file5.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, mock_scope, self.tmp_file5.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file6.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, mock_scope, self.tmp_file6.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file12.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, mock_scope, self.tmp_file12.name, ReplicaState.UNAVAILABLE)

        for replica in replicalist_scope_declarebad:
            suspicious_pfns = replica['rses'][self.rse4suspicious_id]
            #  Declare each file as suspicious multiple times
            for i in range(3):
                print("Declaring suspicious file replica: " + suspicious_pfns[0])
                # The reason must contain the word "checksum", so that the replica can be declared bad.
                replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
                sleep(1)
            if replica['name'] == self.tmp_file7.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_declarebad, self.tmp_file7.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file9.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_declarebad, self.tmp_file9.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file11.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_declarebad, self.tmp_file11.name, ReplicaState.UNAVAILABLE)
            if replica['name'] == self.tmp_file13.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_declarebad, self.tmp_file13.name, ReplicaState.UNAVAILABLE)

        for replica in replicalist_scope_nopolicy:
            suspicious_pfns = replica['rses'][self.rse4suspicious_id]
            #  Declare each file as suspicious multiple times
            for i in range(3):
                print("Declaring suspicious file replica: " + suspicious_pfns[0])
                # The reason must contain the word "checksum", so that the replica can be declared bad.
                replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
                sleep(1)
            if replica['name'] == self.tmp_file8.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_nopolicy, self.tmp_file8.name, ReplicaState.UNAVAILABLE)

        for replica in replicalist_scope_ignore:
            suspicious_pfns = replica['rses'][self.rse4suspicious_id]
            #  Declare each file as suspicious multiple times
            for i in range(3):
                print("Declaring suspicious file replica: " + suspicious_pfns[0])
                # The reason must contain the word "checksum", so that the replica can be declared bad.
                replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
                sleep(1)
            if replica['name'] == self.tmp_file10.name:
                print("Updating replica state as unavailable: " + replica['rses'][self.rse4recovery_id][0])
                update_replica_state(self.rse4recovery_id, self.scope_ignore, self.tmp_file10.name, ReplicaState.UNAVAILABLE)

        # Gather replica info after setting initial replica statuses
        replicalist_scope_mock = list(list_replicas(dids=self.listdids_mock))
        replicalist_scope_declarebad = list(list_replicas(dids=self.listdids_declarebad))
        replicalist_scope_nopolicy = list(list_replicas(dids=self.listdids_nopolicy))
        replicalist_scope_ignore = list(list_replicas(dids=self.listdids_ignore))

        # Checking if the status changes were effective
        for replica in replicalist_scope_mock:
            if replica['name'] == self.tmp_file1.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert replica['states'][self.rse4recovery_id] == 'AVAILABLE'
            if replica['name'] == self.tmp_file2.name:
                assert (self.rse4suspicious_id in replica['states']) is False
                assert replica['states'][self.rse4recovery_id] == 'AVAILABLE'
            if replica['name'] == self.tmp_file3.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file4.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file5.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file6.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file12.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_declarebad:
            if replica['name'] == self.tmp_file7.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file9.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file11.name:
                # tmp_file11 should be ignored, as it doesn't have a datatype
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file13.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_nopolicy:
            if replica['name'] == self.tmp_file8.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_ignore:
            if replica['name'] == self.tmp_file10.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        # Checking if self.tmp_file2 was declared as 'BAD' on rse4suspicious
        self.from_date = datetime.utcnow() - timedelta(days=1)
        bad_replicas_list = list_bad_replicas_status(rse_id=self.rse4suspicious_id, younger_than=self.from_date, vo=vo)
        bad_checklist = [(badf['name'], badf['rse_id'], badf['state']) for badf in bad_replicas_list]

        assert (self.tmp_file1.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file2.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file3.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file4.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file5.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file6.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file7.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file8.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file9.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file10.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file11.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file12.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file13.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist

        bad_replicas_list = list_bad_replicas_status(rse_id=self.rse4recovery_id, younger_than=self.from_date, vo=vo)
        bad_checklist = [(badf['name'], badf['rse_id'], badf['state']) for badf in bad_replicas_list]

        assert (self.tmp_file1.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file2.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file3.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file4.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file5.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file6.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file7.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file8.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file9.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file10.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file11.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file12.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file13.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist

        # Purposefully not checking for the 'SUSPICIOUS' status on rse4suspicious.
        # The only existing function (to date) gathering info about 'SUSPICIOUS' replicas
        # is used (currently only) from the tested replica_recoverer itself.

        # Write policy file to include test policies
        json_data = []
        with open("/opt/rucio/etc/test_replica_recoverer.json", mode="w") as json_file:
            # Any combination of datatype and scope can be given any action.
            # For the test, the actions for json_testentry1 and json_testentry2 have been arbitrarily chosen.
            # Entries that are higher up in the JSON file have priority over those that are below them.
            json_hits = {"action": "declare bad", "datatype": ["HITS"], "scope": []}
            json_raw = {"action": "ignore", "datatype": ["RAW"], "scope": []}
            # json_testentry1 and json_testentry2 need to be above the other entries, otherwise
            # they will always "lose" to an entry with either "datatype": [] or "scope": [].
            # "datatype": [] and "scope": [] are wildcards; they stand for every datatype or scope.
            json_testentry1 = {"action": "ignore", "datatype": ["testtypedeclarebad"], "scope": [str(self.scope_ignore)]}
            json_testentry2 = {"action": "declare bad", "datatype": ["testtypeignore"], "scope": [str(self.scope_declarebad)]}
            json_testentry3 = {"action": "dry run", "datatype": ["testtypedryrun"], "scope": [str(self.scope_declarebad)]}
            json_testentry4 = {"action": "declare bad", "datatype": ["testtypedeclarebad"], "scope": []}
            json_testentry5 = {"action": "ignore", "datatype": ["testtypeignore"], "scope": []}
            json_testentry6 = {"action": "declare bad", "datatype": [], "scope": [str(self.scope_declarebad)]}
            json_testentry7 = {"action": "ignore", "datatype": [], "scope": [str(self.scope_ignore)]}
            json_data.append(json_hits)
            json_data.append(json_raw)
            json_data.append(json_testentry1)
            json_data.append(json_testentry2)
            json_data.append(json_testentry3)
            json_data.append(json_testentry4)
            json_data.append(json_testentry5)
            json_data.append(json_testentry6)
            json_data.append(json_testentry7)
            json.dump(json_data, json_file)

    def test_replica_recoverer(self, vo):
        """ REPLICA RECOVERER: Testing declaration of suspicious replicas as bad if they are found available on other RSEs.

            setUp function (above) is supposed to run first
            (nose does this automatically):

            - uploads 6 test files to two test RSEs (rse4recovery, rse4suspicious)
            - prepares their statuses to be as follows:

        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # Name         State(s) declared on rse4recovery       State(s) declared on rse4suspicious        Scope                     Metadata "datatype"
        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # tmp_file1    available                                suspicious (available)                    mock_scope
        # tmp_file2    available                                suspicious + bad (unavailable)            mock_scope
        # tmp_file3    unavailable                              suspicious (available)                    mock_scope                  RAW
        # tmp_file4    unavailable                              suspicious (available)                    mock_scope                  testtypedeclarebad
        # tmp_file5    unavailable                              suspicious (available)                    mock_scope                  testtypenopolicy
        # tmp_file6    unavailable                              suspicious (available)                    mock_scope                  testtypeignore
        # tmp_file7    unavailable                              suspicious (available)                    scope_declarebad            testtypenopolicy
        # tmp_file8    unavailable                              suspicious (available)                    scope_nopolicy              testtypenopolicy
        # tmp_file9    unavailable                              suspicious (available)                    scope_declarebad            testtypeignore
        # tmp_file10   unavailable                              suspicious (available)                    scope_ignore                testtypedeclarebad
        # tmp_file11   unavailable                              suspicious (available)                    scope_declarebad            <none>
        # tmp_file12   unavailable                              suspicious (available)                    mock_scope                  <none>
        # tmp_file13   unavailable                              suspicious (available)                    scope_declarebad            testtypedryrun
        # ----------------------------------------------------------------------------------------------------------------------------------------------------

            - Explanation: Suspicious replicas that are the last remaining copy (unavailable on rse4recovery) are handled differently depending
                            by their metadata "datatype".
                            - Files that are the last remaining copy, but do not have a data type, are automatically ignored. For this reason, testing
                              just the scope policies (tmp_file7 and tmp_file8) still requires a data type.
                            - RAW files have the poilcy to be ignored.
                            - testtypedeclarebad files are of a fictional type that has the policy of being declared bad.
                            - testtypenopolicy files are of a fictional type that doesn't have a specified policy, meaning they should be ignored by default.
                            - scope_declarebad files belong to a fictional scope that has the policy of being declared bad.
                            - scope_nopolicy files belong to a fictional scope that doesn't have a specified policy, meaning they should be ignored by default.
                            If a policy is set for the file type and the scope, then the policy for the file type takes priority (meaning tmp_file9 should be
                            ignored).

            Runs the Test:

            - running suspicious_replica_recoverer

            Concluding:

            - checks that tmp_file1, tmp_file4, tmp_file7, tmp_file9, tmp_file10 were declared as 'BAD' on rse4suspicious

        """

        try:
            run(once=True, younger_than=1, nattempts=2, limit_suspicious_files_on_rse=11, json_file_name="/opt/rucio/etc/test_replica_recoverer.json", sleep_time=0, active_mode=True)
        except KeyboardInterrupt:
            stop()

        # Checking the outcome:
        # We expect to see four changes: tmp_file1, tmp_file4, tmp_file7, tmp_file9, tmp_file10 should be declared as bad on rse4suspicious
        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # Name         State(s) declared on rse4recovery       State(s) declared on rse4suspicious        Scope                     Metadata "datatype"
        # ----------------------------------------------------------------------------------------------------------------------------------------------------
        # tmp_file1    available                                suspicious + bad (unavailable)            mock_scope                  <none>
        # tmp_file2    available                                suspicious + bad (unavailable)            mock_scope                  <none>
        # tmp_file3    unavailable                              suspicious (available)                    mock_scope                  RAW
        # tmp_file4    unavailable                              suspicious + bad (unavailable)            mock_scope                  testtypedeclarebad
        # tmp_file5    unavailable                              suspicious (available)                    mock_scope                  testtypenopolicy
        # tmp_file6    unavailable                              suspicious (available)                    mock_scope                  testtypeignore
        # tmp_file7    unavailable                              suspicious + bad (unavailable)            scope_declarebad            testtypenopolicy
        # tmp_file8    unavailable                              suspicious (available)                    scope_nopolicy              testtypenopolicy
        # tmp_file9    unavailable                              suspicious + bad (unavailable)            scope_declarebad            testtypeignore
        # tmp_file10   unavailable                              suspicious (available)                    scope_ignore                testtypedeclarebad
        # tmp_file11   unavailable                              suspicious (available)                    scope_declarebad            <none>
        # tmp_file12   unavailable                              suspicious (available)                    mock_scope                  <none>
        # tmp_file13   unavailable                              suspicious (available)                    scope_declarebad            testtypedryrun
        # ----------------------------------------------------------------------------------------------------------------------------------------------------

        # Gather replica info after replica_recoverer has run.
        replicalist_scope_mock = list(list_replicas(dids=self.listdids_mock))
        replicalist_scope_declarebad = list(list_replicas(dids=self.listdids_declarebad))
        replicalist_scope_nopolicy = list(list_replicas(dids=self.listdids_nopolicy))
        replicalist_scope_ignore = list(list_replicas(dids=self.listdids_ignore))

        for replica in replicalist_scope_mock:
            if replica['name'] == self.tmp_file1.name or replica['name'] == self.tmp_file2.name:
                assert (self.rse4suspicious_id in replica['states']) is False
                assert replica['states'][self.rse4recovery_id] == 'AVAILABLE'
            if replica['name'] == self.tmp_file3.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file4.name:
                # The 'states' key doesn't exist if the replica isn't available on at least one RSE
                assert not replica.get('states')
            if replica['name'] == self.tmp_file5.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file12.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_declarebad:
            if replica['name'] == self.tmp_file7.name:
                # The 'states' key doesn't exist if the replica isn't available on at least one RSE
                assert not replica.get('states')
            if replica['name'] == self.tmp_file9.name:
                assert not replica.get('states')
            if replica['name'] == self.tmp_file11.name:
                # tmp_file11 should have been ignored, as it doesn't have a datatype
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False
            if replica['name'] == self.tmp_file13.name:
                # tmp_file13 should have been ignored, as it is running as a dry run
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_nopolicy:
            if replica['name'] == self.tmp_file8.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        for replica in replicalist_scope_ignore:
            if replica['name'] == self.tmp_file10.name:
                assert replica['states'][self.rse4suspicious_id] == 'AVAILABLE'
                assert (self.rse4recovery_id in replica['states']) is False

        # Checking if replicas were declared as 'BAD'
        bad_replicas_list = list_bad_replicas_status(rse_id=self.rse4suspicious_id, younger_than=self.from_date, vo=vo)
        bad_checklist = [(badf['name'], badf['rse_id'], badf['state']) for badf in bad_replicas_list]

        assert (self.tmp_file1.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file2.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file3.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file4.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file5.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file6.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file7.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file8.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file9.name, self.rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist
        assert (self.tmp_file10.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file11.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file12.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file13.name, self.rse4suspicious_id, BadFilesStatus.BAD) not in bad_checklist

        bad_replicas_list = list_bad_replicas_status(rse_id=self.rse4recovery_id, younger_than=self.from_date, vo=vo)
        bad_checklist = [(badf['name'], badf['rse_id'], badf['state']) for badf in bad_replicas_list]

        assert (self.tmp_file1.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file2.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file3.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file4.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file5.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file6.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file7.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file8.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file9.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file10.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file11.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file12.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist
        assert (self.tmp_file13.name, self.rse4recovery_id, BadFilesStatus.BAD) not in bad_checklist


@pytest.mark.parametrize(
    "file_config_mock", [{"overrides": [
        ("replicarecoverer", "rule_rse_expression", "type=SCRATCHDISK"),
        ('replicarecoverer', 'use_file_metadata', 'False'),
        ('replicarecoverer', 'did_name_expression', '^.*')]}],
    indirect=True)
def test_vo_agnostic_rules(file_config_mock, replica_client, rse_factory, scope_factory, file_factory, vo):
    """
    Runs only a single time, make sure it makes a new rule with this config.

    Examining files without a datatype, but because the "use_file_metadata" is turned off,
    files without a datatype can be acted on.
    """
    rse4suspicious, rse4suspicious_id = rse_factory.make_posix_rse(deterministic=True, vo=vo)
    rse4recovery, rse4recovery_id = rse_factory.make_posix_rse(deterministic=True, vo=vo)

    add_rse_attribute(rse4suspicious_id, "enable_suspicious_file_recovery", True)
    add_rse_attribute(rse4recovery_id, "enable_suspicious_file_recovery", True)

    # Create new scopes
    _, [scope_declarebad] = scope_factory(vos=[vo])

    tmp_file_delare_bad = file_factory.file_generator()

    replicas_without_types = [{'scope': scope_declarebad, 'name': tmp_file_delare_bad.name, 'type': DIDType.FILE}]
    for rse in [rse4suspicious, rse4recovery]:
        cmd = f'rucio -v upload --legacy --rse {rse} --scope {scope_declarebad} {tmp_file_delare_bad}'
        exitcode, _, _ = execute(cmd)
        assert exitcode == 0

    remove(tmp_file_delare_bad)
    rse_expression_parser.REGION.invalidate()

    update_replica_state(rse4recovery_id, scope_declarebad, tmp_file_delare_bad.name, ReplicaState.UNAVAILABLE)
    replical_declare_bad = list(list_replicas(dids=replicas_without_types))

    for replica in replical_declare_bad:
        suspicious_pfns = replica['rses'][rse4suspicious_id]
        for _ in range(3):
            # The reason must contain the word "checksum", so that the replica can be declared bad.
            replica_client.declare_suspicious_file_replicas([suspicious_pfns[0], ], 'checksum')
            sleep(1)

    recovery_policy = [{"action": "declare bad", "datatype": [str(tmp_file_delare_bad.name)], "scope": [str(scope_declarebad)]}]
    recovery_policy_json = "/opt/rucio/etc/test_replica_recoverer_vo_agonistism.json"
    json.dump(recovery_policy, open(recovery_policy_json, mode="w"))

    try:
        run(
            once=True,
            younger_than=1,
            nattempts=2,
            limit_suspicious_files_on_rse=2,
            json_file_name=recovery_policy_json,
            sleep_time=0,
            active_mode=True
        )
    except KeyboardInterrupt:
        stop()

    # dids without types now have a different behavior

    for replica in replicas_without_types:
        # Files that are now acted on now the datatype can be found.
        if replica['name'] == tmp_file_delare_bad.name:
            # tmp_file11 should no longer be ignored.
            assert not replica.get('states')

    bad_replicas_list = list_bad_replicas_status(rse_id=rse4suspicious_id, vo=vo)
    bad_checklist = [(badf['name'], badf['rse_id'], badf['state']) for badf in bad_replicas_list]
    assert (tmp_file_delare_bad.name, rse4suspicious_id, BadFilesStatus.BAD) in bad_checklist




'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '37.0.0rc2',
    'branch_nick': 'master',
    'revision_id': '6fb393d63e8f50733e00ce0b0234380ac12cd607',
    'revno': 13608
}
